
#include "php_htmltmplpro.h"
#include <htmltmplpro/tmplpro.h>
#include <stddef.h>

#if (PHP_MAJOR_VERSION < 5) || (PHP_MAJOR_VERSION == 5 && PHP_MINOR_VERSION < 1) 
#error HTMLTMPLPRO_EXT_NAME " extension require php5.1 or higher."
#endif

/* If you declare any globals in php_htmltmplpro.h uncomment this:
ZEND_DECLARE_MODULE_GLOBALS(htmltmplpro)
*/

/* True global resources - no need for thread safety here */
//static int le_htmltmplpro;

/* {{{ htmltmplpro_functions[]
 *
 * Every user visible function must have an entry in htmltmplpro_functions[].
 */
zend_function_entry htmltmplpro_functions[] = {
	PHP_FE(htmltmplpro_versioninfo,	NULL)		/* For testing, remove later. */
	{NULL, NULL, NULL}	/* Must be the last line in htmltmplpro_functions[] */
};
/* }}} */

/* {{{ htmltmplpro_module_entry
 */
zend_module_entry htmltmplpro_module_entry = {
#if ZEND_MODULE_API_NO >= 20010901
	STANDARD_MODULE_HEADER,
#endif
	HTMLTMPLPRO_EXT_NAME,
	htmltmplpro_functions,
	PHP_MINIT(htmltmplpro),
	PHP_MSHUTDOWN(htmltmplpro),
   	NULL,	/* PHP_RINIT(htmltmplpro),		/* Replace with NULL if there's nothing to do at request start */
	NULL,	/* PHP_RSHUTDOWN(htmltmplpro),	/* Replace with NULL if there's nothing to do at request end */
	PHP_MINFO(htmltmplpro),
#if ZEND_MODULE_API_NO >= 20010901
	HTMLTMPLPRO_EXT_VERSION, /* Replace with version number for your extension */
#endif
	STANDARD_MODULE_PROPERTIES
};
/* }}} */

#ifdef COMPILE_DL_HTMLTMPLPRO
ZEND_GET_MODULE(htmltmplpro)
#endif

/* {{{ PHP_INI
 */
/* Remove comments and fill if you need to have entries in php.ini
PHP_INI_BEGIN()
    STD_PHP_INI_ENTRY("htmltmplpro.global_value",      "42", PHP_INI_ALL, OnUpdateLong, global_value, zend_htmltmplpro_globals, htmltmplpro_globals)
    STD_PHP_INI_ENTRY("htmltmplpro.global_string", "foobar", PHP_INI_ALL, OnUpdateString, global_string, zend_htmltmplpro_globals, htmltmplpro_globals)
PHP_INI_END()
*/
/* }}} */

/* {{{ php_htmltmplpro_init_globals
 */
/* Uncomment this function if you have INI entries
static void php_htmltmplpro_init_globals(zend_htmltmplpro_globals *htmltmplpro_globals)
{
	htmltmplpro_globals->global_value = 0;
	htmltmplpro_globals->global_string = NULL;
}
*/
/* }}} */

/* {{{ PHP_MSHUTDOWN_FUNCTION
 */
PHP_MSHUTDOWN_FUNCTION(htmltmplpro)
{
	/* uncomment this line if you have INI entries
	UNREGISTER_INI_ENTRIES();
	*/
	tmplpro_procore_done();
	return SUCCESS;
}
/* }}} */

/* Remove if there's nothing to do at request start */
/* {{{ PHP_RINIT_FUNCTION
 */
PHP_RINIT_FUNCTION(htmltmplpro)
{
	return SUCCESS;
}
/* }}} */

/* Remove if there's nothing to do at request end */
/* {{{ PHP_RSHUTDOWN_FUNCTION
 */
PHP_RSHUTDOWN_FUNCTION(htmltmplpro)
{
	return SUCCESS;
}
/* }}} */

/* {{{ PHP_MINFO_FUNCTION
 */
PHP_MINFO_FUNCTION(htmltmplpro)
{
	php_info_print_table_start();
	php_info_print_table_header(2, "htmltmplpro support", "enabled");
	php_info_print_table_row(2, "Version", HTMLTMPLPRO_EXT_VERSION);
	php_info_print_table_end();

	/* Remove comments if you have entries in php.ini
	DISPLAY_INI_ENTRIES();
	*/
}
/* }}} */


/*----------------------------------------------------------------

/* Resource struct, not the object :) */
/* Extends zend object */
typedef struct _ze_htp_object {
	zend_object zo;
	struct tmplpro_param* htp_ptr;
	zval* userfuncmap;
	zval* input_zval;
	zval* path_zarray_arg;
	char** path;
	char* method_loop_init;
	char* method_loop_next;
	char* method_loop_count;
	size_t method_loop_init_len;
	size_t method_loop_next_len;
	size_t method_loop_count_len;
} ze_htp_object;

static zend_class_entry *htmltmplpro_class_entry_ce;

/* {{{ HTP_DECLARE_OBJECTS */
#define HTP_DECLARE_OBJECTS \
	struct tmplpro_param *htp_obj; \
	ze_htp_object *ze_obj; \
	zval *php_this = getThis();
/* }}} */

/* {{{ HTP_DECLARE_OBJECTS */
#define HTP_INIT_OBJECTS \
	if (!php_this) { \
		php_error(E_WARNING, "this function can only be invoked as a method of HTML Template Pro object"); \
		RETURN_FALSE; \
	} else { \
		ze_obj = (ze_htp_object*) zend_object_store_get_object(php_this TSRMLS_CC); \
		htp_obj = ze_obj->htp_ptr; \
		if (!htp_obj) { \
			php_error(E_ERROR, "Internal Error: Invalid HTML Template Pro object"); \
			RETURN_FALSE; \
		} \
	}
/* }}} */


#define HTP_ZVAL_FIELD_CLEAR(ze_obj, input_zval) { \
	if (ze_obj->input_zval!=NULL) { \
		zval_ptr_dtor(&ze_obj->input_zval);	\
		ze_obj->input_zval=(zval*) NULL; \
	} \
}

#define HTP_ZVAL_FIELD_SET(ze_obj, input_zval, val) { \
	HTP_ZVAL_FIELD_CLEAR(ze_obj, input_zval);	  \
	if (val!=NULL) { \
		Z_ADDREF_P(val); \
		ze_obj->input_zval=val; \
	} \
}

static void _free_option_path (ze_htp_object *ze_obj) {
	if (ze_obj->path!=NULL) {
		efree(ze_obj->path);
		ze_obj->path=NULL;
	}
}

static void _alloc_option_path (ze_htp_object *ze_obj, long count) {
	ze_obj->path = ecalloc (sizeof(char*), count+1);
}

static void _push_option_path (ze_htp_object *ze_obj, zval* val, long pos) {
	(ze_obj->path)[pos] = Z_STRVAL_P(val);
	(ze_obj->path)[pos+1] = NULL;
	//php_printf("pushed path [%s](%p) at pos %ld\n", Z_STRVAL_P(val),Z_STRVAL_P(val),pos);
}


/* ######################################################################### */

static int fill_scope_stack(struct tmplpro_param *htp_obj, long argument_count TSRMLS_DC)
{
	zval **zargp;
	zval *** args;
	HashTable *arr_hash;
    int array_count;
	long i;
	int ztype;

    args = (zval ***)safe_emalloc(argument_count, sizeof(zval **), 0);
	if (argument_count == 0 || 
        zend_get_parameters_array_ex(argument_count, args) == FAILURE) { 
        efree(args); 
		php_error(E_WARNING, "output and print methods: not enough arguments. At least 1 should be given.");
        return 1; 
    } 
    for (i=0; i<argument_count; i++) {
		ztype = Z_TYPE_PP(args[i]);
        if (ztype != IS_ARRAY && ztype != IS_OBJECT) {
			//php_printf("not an array");
			efree(args);
			php_error(E_WARNING, "output and print methods only accept params as (map [i.e. associative array or object with properties] | numeric array of maps [, map, map, ...])");
			return 1;
		}
    }

	ztype = Z_TYPE_PP(args[0]);
	if (ztype == IS_ARRAY) {
		arr_hash = Z_ARRVAL_PP(args[0]);
		array_count = zend_hash_num_elements(arr_hash);

		tmplpro_clear_option_param_map(htp_obj);
		if (array_count==0) {
			tmplpro_push_option_param_map(htp_obj, *(args[0]), 0);
		} else {
			for (i=0; i<array_count; i++) { 
				if (zend_hash_index_find(arr_hash, i, (void **) &zargp) == FAILURE || 
					zargp == NULL ||
					Z_TYPE_PP(zargp)!= IS_ARRAY) {
					/* first arg seems not to be a numeric array -- let's push it as is. */
					tmplpro_clear_option_param_map(htp_obj);
					tmplpro_push_option_param_map(htp_obj, *(args[0]), 0);
					break;
				} else {
					tmplpro_push_option_param_map(htp_obj,*zargp, 0);
				}
			}
		}
	} else if (ztype == IS_OBJECT) {
		tmplpro_clear_option_param_map(htp_obj);
		tmplpro_push_option_param_map(htp_obj, *(args[0]), 0);
	}

	/* pushing the rest of supplied roots */
    for (i=1; i<argument_count; i++) { 
		tmplpro_push_option_param_map(htp_obj, *(args[i]), 0);
	}

    efree(args); 
	return 0;
}

/* {{{ proto string HTP->output(numeric array of maps | map [, map, map, ...]);
   Map is either associative array or an object.
   Note that the root variable scope should be the last supplied associative array.

   Process template and return the output as a string.
*/
static PHP_METHOD(htmltmplpro,output) 
{
	long argument_count = ZEND_NUM_ARGS();
	PSTRING retstr;
	int retval = 0;
	HTP_DECLARE_OBJECTS;
	HTP_INIT_OBJECTS;
	if (fill_scope_stack(htp_obj, argument_count TSRMLS_CC)) WRONG_PARAM_COUNT;
	retstr = tmplpro_tmpl2pstring(htp_obj, &retval);
	RETURN_STRINGL((char*)retstr.begin, retstr.endnext-retstr.begin, 1);
}
/* }}} */

/* {{{ proto void HTP->print(numeric array of maps | map [, map, map, ...]) 
   Map is either associative array or an object.
   Note that the root variable scope should be the last supplied associative array.

   Process template and print the output to php output stream.
*
*/
static PHP_METHOD(htmltmplpro,print)
{
	long argument_count = ZEND_NUM_ARGS();
	HTP_DECLARE_OBJECTS;
	HTP_INIT_OBJECTS;
	if (fill_scope_stack(htp_obj, argument_count TSRMLS_CC)) WRONG_PARAM_COUNT;
	RETURN_LONG(tmplpro_exec_tmpl(htp_obj));
}
/* }}} */


/* {{{ proto HTP->setInputFile(string filename)
   Set the input template to be a given file.
*/
static PHP_FUNCTION(htmltmplpro_set_input_file)
{
	zval *zarg;
	long argument_count = ZEND_NUM_ARGS();
	HTP_DECLARE_OBJECTS;
	//php_printf("set_input_file: got %ld args\n", argument_count);

	HTP_INIT_OBJECTS;
	if (argument_count == 0) {
		tmplpro_set_option_filename(htp_obj, NULL);
		HTP_ZVAL_FIELD_CLEAR(ze_obj, input_zval);
	} else {
		if (zend_parse_parameters(argument_count TSRMLS_CC, "z", &zarg) == FAILURE) {
			WRONG_PARAM_COUNT;
		}
		if (Z_TYPE_P(zarg)==IS_NULL) {
			tmplpro_set_option_filename(htp_obj, NULL);
			HTP_ZVAL_FIELD_CLEAR(ze_obj, input_zval);
		} else {
			convert_to_string(zarg);
			tmplpro_set_option_filename(htp_obj, Z_STRVAL_P(zarg));
			HTP_ZVAL_FIELD_SET(ze_obj, input_zval, zarg);
		}
	}
	RETURN_TRUE;
}
/* }}} */

/* {{{ proto HTP->setInputString(string str)
   Set the input template to be a given string.
*/
static PHP_FUNCTION(htmltmplpro_set_input_string)
{
	zval *zarg;
	char *name;
    int name_len;
	PSTRING pstr = {NULL, NULL};
	long argument_count = ZEND_NUM_ARGS();
	HTP_DECLARE_OBJECTS;
	//php_printf("set_input_string: got %ld args\n", argument_count);

	HTP_INIT_OBJECTS;
	if (argument_count == 0) {
		tmplpro_set_option_scalarref(htp_obj, pstr);
		HTP_ZVAL_FIELD_CLEAR(ze_obj, input_zval);
	} else {
		if (zend_parse_parameters(argument_count TSRMLS_CC, "z", &zarg) == FAILURE) {
			WRONG_PARAM_COUNT;
		}
		if (Z_TYPE_P(zarg)==IS_NULL) {
			tmplpro_set_option_scalarref(htp_obj, pstr);
			HTP_ZVAL_FIELD_CLEAR(ze_obj, input_zval);
		} else {
			convert_to_string(zarg);
			name = Z_STRVAL_P(zarg);
			name_len = Z_STRLEN_P(zarg);
			pstr.begin = name;
			pstr.endnext = name + name_len;
			tmplpro_set_option_scalarref(htp_obj, pstr);
			HTP_ZVAL_FIELD_SET(ze_obj, input_zval, zarg);
		}
	}
	RETURN_TRUE;
}
/* }}} */

/* {{{ proto string HTP->getInputString()
   if input template was set as a string returns it,
   otherwise returns php null.
*/
static PHP_FUNCTION(htmltmplpro_get_input_string)
{
	zval *zarg;
	char *name;
    int name_len;
	PSTRING pstr = {NULL, NULL};
	long argument_count = ZEND_NUM_ARGS();
	HTP_DECLARE_OBJECTS;
	HTP_INIT_OBJECTS;
	if (argument_count != 0) {
		WRONG_PARAM_COUNT;
	} else {
		PSTRING pinput = tmplpro_get_option_scalarref(htp_obj);
		if (pinput.begin!=NULL && ze_obj->input_zval!=NULL) {
			zval_dtor(return_value);
			*return_value = *(ze_obj->input_zval);
			zval_copy_ctor(return_value);
		}
	}
}
/* }}} */

/* {{{ proto zval HTP->getInputFile()
   if filename for input template was set returns it,
   otherwise returns php null.
*/
static PHP_FUNCTION(htmltmplpro_get_input_file)
{
	zval *zarg;
	char *name;
    int name_len;
	PSTRING pstr = {NULL, NULL};
	long argument_count = ZEND_NUM_ARGS();
	HTP_DECLARE_OBJECTS;
	HTP_INIT_OBJECTS;
	if (argument_count != 0) {
		WRONG_PARAM_COUNT;
	} else {
		if (tmplpro_get_option_filename(htp_obj)!=NULL && ze_obj->input_zval!=NULL) {
			zval_dtor(return_value);
			*return_value = *(ze_obj->input_zval);
			zval_copy_ctor(return_value);
		}
	}
}
/* }}} */

/* {{{ proto zval HTP->reset()
   reset userspace configuration options to the default values.
   Useful to reinitialize existing object instead of creation anew.
*/
static PHP_METHOD(htmltmplpro,reset)
{
	zval *zarg;
	char *name;
    int name_len;
	PSTRING pstr = {NULL, NULL};
	long argument_count = ZEND_NUM_ARGS();
	HTP_DECLARE_OBJECTS;
	HTP_INIT_OBJECTS;
	if (argument_count != 0) {
		WRONG_PARAM_COUNT;
	} else {
		tmplpro_reset_int_options(htp_obj);
		tmplpro_set_option_template_root(htp_obj, NULL);
		/* TODO: path, ... ? */
	}
}
/* }}} */


#define STR_EQ(A,B,BLEN) (sizeof(A)-1 == BLEN && 0==strncmp(B,A,BLEN))

/* {{{ proto HTP->setopt(string optname, zval value);
   optname is the library option, such as "path", "strict", "no_includes", etc.
   For all library options and what they means please refer to the documentation
   of perl module HTML::Template::Pro. Future versions of libhtmltmplpro 
   documentation will include this information as well.

   generally the value to be set is an integer. For booleans it should be 0 or 1.
   "path" option accepts a numeric array of paths,

   there is a php-specific "functions" option that accept the associative array 
   (map) of "template function name" => "php function name" pairs
   that is used to resolve calls to functions in template to user provided php 
   functions.
*/
static PHP_METHOD(htmltmplpro,setopt)
{
	const char *optname;
	int optnamelen;
	zval *zarg=NULL;
	long argument_count = ZEND_NUM_ARGS();
	int exitcode;
	zval ** *parameter_array;

	HTP_DECLARE_OBJECTS;
	HTP_INIT_OBJECTS;

	if(argument_count == 0) {
		php_error(E_ERROR, "setopt: option name is required");
		WRONG_PARAM_COUNT;
	}
	
	/* argument count is correct, now retrieve arguments */
	parameter_array = (zval***) ecalloc(argument_count, sizeof(zval **));

	if(zend_get_parameters_array_ex(argument_count, parameter_array) != SUCCESS)
		WRONG_PARAM_COUNT;
	if (parameter_array==NULL) WRONG_PARAM_COUNT;
	if (argument_count>2) {
		php_error(E_ERROR, "setopt method: does not support more than 2 arguments");
		goto exit_setopt;
	}
	if (!IS_STRING==Z_TYPE_PP(parameter_array[0])) {
		php_error(E_ERROR, "setopt method: 1st argument (option name) should be a string");
		goto exit_setopt;
	}
	optname = Z_STRVAL_PP(parameter_array[0]);
	optnamelen = Z_STRLEN_PP(parameter_array[0]);
	if (argument_count > 1) {
		zarg = *(parameter_array[1]);
	}

	if (STR_EQ("functions",optname,optnamelen)) {
		if (zarg==NULL || Z_TYPE_P(zarg)==IS_NULL) {
			tmplpro_set_option_expr_func_map(htp_obj, NULL);
			HTP_ZVAL_FIELD_CLEAR(ze_obj, userfuncmap);
		} else if (Z_TYPE_P(zarg)==IS_ARRAY) {
			tmplpro_set_option_expr_func_map(htp_obj, Z_ARRVAL_P(zarg));
			HTP_ZVAL_FIELD_SET(ze_obj, userfuncmap, zarg);
		} else {
			php_error(E_ERROR, "setopt method: 2nd argument for functions option should be an array");
		}
		goto exit_setopt;
	} else if (STR_EQ("path",optname,optnamelen)) {
		zval** zvalp;
		_free_option_path(ze_obj);
		HTP_ZVAL_FIELD_CLEAR(ze_obj, path_zarray_arg);
		tmplpro_set_option_path(htp_obj, NULL);
		if (zarg==NULL || Z_TYPE_P(zarg)==IS_NULL) {
			// noargs; just clear path
			goto  exit_setopt;
		}
		if (IS_ARRAY!=Z_TYPE_P(zarg)) {
			php_error(E_ERROR, "setopt method: 2nd argument for path option should be an array");
			goto exit_setopt;
		} else {
			HashTable *arr_hash = Z_ARRVAL_P(zarg);
			long array_count = zend_hash_num_elements(arr_hash);
			if (array_count == 0) {
				// noargs; just clear path
				goto  exit_setopt;
			} else {
				long i = 0;
				int is_assoc = 0;
				HTP_ZVAL_FIELD_SET(ze_obj, path_zarray_arg, zarg);
				_alloc_option_path(ze_obj, array_count);
				for (i=0; i<array_count; i++) { 
					//php_printf("set path: try %ld array_count =%ld\n", i, array_count);
					if (zend_hash_index_find(arr_hash, i, (void **) &zvalp) != SUCCESS || 
						zvalp == NULL) {
						is_assoc = 1; break;
					} else {
						if (IS_STRING != Z_TYPE_PP(zvalp)) {
							convert_to_string_ex(zvalp);
							php_error(E_WARNING, "setopt(path): non-string value %s explicitly converted to string", Z_STRVAL_PP(zvalp));
						}
						_push_option_path(ze_obj, *zvalp, i);
					}
				}
				if (is_assoc) {
					php_error(E_WARNING, "setopt(path): array argument seems to be an associative array, but numeric array is required");
				} else {
					tmplpro_set_option_path(htp_obj, ze_obj->path);
				}
			}
		}
		/*{
			char** cpath = ze_obj->path;
			php_printf("setopt info: path:");
			while (*cpath != NULL) {
				php_printf("[%s](%p) ", *cpath,*cpath);
				cpath++;
			}
			php_printf("\n");
			}*/
	} else {
		long int longval = 0;
		if (zarg!=NULL) {
			SEPARATE_ZVAL(&zarg);
			convert_to_long(zarg);
			longval = Z_LVAL_P(zarg);
		}
		exitcode = tmplpro_set_int_option(htp_obj, optname, longval);
		if (exitcode!=0) {
			php_error(E_WARNING, "setopt method: unknown option %s or invalid value", optname);
		}
	}
	exit_setopt:;
	efree(parameter_array);
	return;
}
/* }}} */

#define GETOPT_RET_ZVAL(conststr, optname,optnamelen, ze_field) \
	if (STR_EQ(conststr,optname,optnamelen)) { \
		if ((ze_field)!=NULL) { \
			zval_dtor(return_value); \
			*return_value = *(ze_field); \
			zval_copy_ctor(return_value); \
		} \
		return; \
	}

/* {{{ proto zval $tmpl->getopt(string optname)
   optname is the library option, such as "path", "strict", "no_includes", etc.
   For all library options and what they means please refer to the documentation
   of perl module HTML::Template::Pro. Future versions of libhtmltmplpro 
   documentation will include this information as well.
   there is also a php-specific "functions" option. See setopt for details.
*/
static PHP_METHOD(htmltmplpro,getopt) 
{
	long argument_count = ZEND_NUM_ARGS();
	char* optname;
	int optnamelen;
	int exitcode;
	int retval;
	HTP_DECLARE_OBJECTS;
	HTP_INIT_OBJECTS;
	if (zend_parse_parameters(argument_count TSRMLS_CC, "s", &optname, &optnamelen) == FAILURE) {
		WRONG_PARAM_COUNT;
	}
	//php_printf("getopt: %s, len=%d\n",optname, optnamelen);
	GETOPT_RET_ZVAL("functions",optname,optnamelen,ze_obj->userfuncmap);
	GETOPT_RET_ZVAL("path",optname,optnamelen,ze_obj->path_zarray_arg);
	retval = tmplpro_get_int_option(htp_obj, optname, &exitcode);
	if (exitcode!=0) {
		php_error(E_WARNING, "getopt method: unknown option %s", optname);
	}
	RETURN_LONG(retval);
}
/* }}} */

/* {{{ proto HTP->error()
   returns exit status of output,print,getopt and setopt methods.
   if those method succeeded it returns 0
   otherwise it returns non-zero exit code.
*/
static PHP_FUNCTION(htmltmplpro_last_error)
{
	long argument_count = ZEND_NUM_ARGS();
	HTP_DECLARE_OBJECTS;
	HTP_INIT_OBJECTS;
	RETURN_LONG(tmplpro_errno(htp_obj));
}
/* }}} */


/* =================== CALLBACKS =========================== */

int php_sample_print_zval_and_key(zval **val, int num_args, va_list args, zend_hash_key *hash_key) 
{ 
    /* Duplicate the zval so that 
     * the original's contents are not destroyed */ 
    zval tmpcopy = **val; 
    /* tsrm_ls is needed by output functions */ 
    TSRMLS_FETCH(); 
    zval_copy_ctor(&tmpcopy); 
    /* Reset refcount & Convert */ 
    INIT_PZVAL(&tmpcopy); 
    convert_to_string(&tmpcopy); 
    /* Output */ 
    php_printf("The value of "); 
    if (hash_key->nKeyLength) { 
        /* String Key / Associative */ 
        PHPWRITE(hash_key->arKey, hash_key->nKeyLength); 
        php_printf(" [associative with key len=%d]", hash_key->nKeyLength); 
    } else { 
        /* Numeric Key */ 
        php_printf("%ld", hash_key->h); 
    } 
    php_printf(" is: "); 
    PHPWRITE(Z_STRVAL(tmpcopy), Z_STRLEN(tmpcopy)); 
    php_printf("\n"); 
    /* Toss out old copy */ 
    zval_dtor(&tmpcopy); 
    /* continue; */ 
    return ZEND_HASH_APPLY_KEEP; 
}

// return 1 if method exists (and could be called successfully?)
static int test_and_call_zend_method_with_0_params (zval* zobj, char* method_name, size_t method_name_len, zval** zretval) {
	if (method_name
		&& Z_OBJ_HT_P(zobj)->get_method != NULL
		&& (Z_OBJ_HT_P(zobj)->get_method(&zobj, method_name, method_name_len TSRMLS_CC)) != NULL
		) {
#if (PHP_MAJOR_VERSION < 5) || (PHP_MAJOR_VERSION == 5 && PHP_MINOR_VERSION < 3) 
		zend_call_method(&zobj, Z_OBJCE_P(zobj), NULL, method_name, method_name_len, zretval, 0, NULL, NULL TSRMLS_CC);
#else
		zend_call_method_with_0_params(&zobj, Z_OBJCE_P(zobj), NULL, method_name, zretval);
#endif
		return 1;
	}
	return 0;
}

#define COUNT_OF_END_STRING_ZERO 1

static ABSTRACT_VALUE* get_ABSTRACT_VALUE_func_impl (ABSTRACT_DATASTATE* no, ABSTRACT_MAP* map, PSTRING name) {
	zval ** mapvalptr;
	ABSTRACT_VALUE* retval = NULL;
	size_t keylen = name.endnext-name.begin;
	char* tmpkey;
	HashTable* ht;
	zval* zvalue = NULL;
	zval tmpzval;
	zend_class_entry *ce;
	if (keylen==0) return NULL;
	tmpkey = estrndup(name.begin, keylen);

	switch (Z_TYPE_P((zval*)map)) {
    case IS_ARRAY:
		ht = (HashTable*) Z_ARRVAL_P((zval*)map);
		//php_printf("get_ABSTRACT_VALUE:in array: asked %p for %.*s\n",map,(int)(keylen), name.begin);
		//zend_hash_apply_with_arguments(ht, php_sample_print_zval_and_key, 0);
		if(zend_hash_find(ht, (char*)tmpkey, keylen+ COUNT_OF_END_STRING_ZERO, (void **)&mapvalptr) == FAILURE) {
			//php_printf("get_ABSTRACT_VALUE: FAILURE\n");
		} else {
			if (NULL!= mapvalptr && NULL!= *mapvalptr) {
				retval = (ABSTRACT_VALUE*) *mapvalptr;
				//php_printf("get_ABSTRACT_VALUE: return %p\n", retval);
			}
		}
		break;
	case IS_OBJECT:
		ce = Z_OBJCE_P((zval*)map);
		ZVAL_STRINGL(&tmpzval,tmpkey,keylen,0);
#if (PHP_MAJOR_VERSION > 5) || (PHP_MAJOR_VERSION == 5 && PHP_MINOR_VERSION >= 3)
		if (instanceof_function(ce, zend_ce_arrayaccess TSRMLS_CC)) {
			if (Z_OBJ_HT_P((zval*)map)->read_dimension) {
				zvalue = Z_OBJ_HT_P((zval*)map)->read_dimension((zval*)map, &tmpzval, BP_VAR_IS TSRMLS_CC);
			} else {
				php_error(E_WARNING, "get_ABSTRACT_VALUE_func_impl: internal error: can't find read_dimension handler");
			}
		} else
#endif
		{
			zvalue = zend_read_property(ce, (zval*)map, (char*)tmpkey, keylen, BP_VAR_IS TSRMLS_CC);
			//php_printf("get_ABSTRACT_VALUE:in object: asked %p for %.*s\n",map,(int)(keylen), name.begin);
			if (zvalue != EG(uninitialized_zval_ptr)) {
				retval = (ABSTRACT_VALUE*) zvalue;
			}
		}
		break;
	default:
		php_error(E_WARNING, "get_ABSTRACT_VALUE_func_impl: internal error: got strange zval");
	}
	efree(tmpkey);
	return retval;
}

static ABSTRACT_ARRAY* ABSTRACT_VALUE2ABSTRACT_ARRAY_func_impl (ABSTRACT_DATASTATE* ze_self, ABSTRACT_VALUE* val) {
	int ztype;
	zval* zvalue;
	if (val==NULL) return NULL;
	ztype = Z_TYPE_P((zval*)val);
	if (ztype == IS_ARRAY ||
		((ztype == IS_OBJECT) &&
		 test_and_call_zend_method_with_0_params((zval*)val, ((struct _ze_htp_object*) ze_self)->method_loop_init, ((struct _ze_htp_object*) ze_self)->method_loop_init_len, &zvalue))
		) {
		return (ABSTRACT_ARRAY*) val;
	} else {
		return NULL;
	}
}

static ABSTRACT_MAP* get_ABSTRACT_MAP_func_impl (ABSTRACT_DATASTATE* ze_self, ABSTRACT_ARRAY* arr,int pos) {
	zval** mapvalptr;
	HashTable* ht;
	zval* zvalue = NULL;
	switch (Z_TYPE_P((zval*)arr)) {
    case IS_ARRAY:
		ht = Z_ARRVAL_P((zval*)arr);
		if(zend_hash_index_find(ht, pos, (void **)&mapvalptr) == FAILURE) {
			return NULL;
		}
		if (NULL== mapvalptr || NULL== *mapvalptr) return NULL;
		zvalue=*mapvalptr;
		break;
	case IS_OBJECT:
		if (!test_and_call_zend_method_with_0_params((zval*)arr, ((struct _ze_htp_object*) ze_self)->method_loop_next, ((struct _ze_htp_object*) ze_self)->method_loop_next_len, &zvalue)) {
			php_error(E_WARNING, "get_ABSTRACT_MAP_func_impl: object is in loop context, but does not implement method %s::%s", (Z_OBJCE_P(((zval*)arr)))->name, ((struct _ze_htp_object*) ze_self)->method_loop_next);
		}
		break;
	default:
		php_error(E_WARNING, "get_ABSTRACT_MAP_func_impl: internal error: got strange zval");
		return NULL;
	}

    if (zvalue == NULL) return NULL;

	switch (Z_TYPE_P(zvalue)) {
    case IS_NULL:
		/* end of loop */
		return NULL;
    case IS_BOOL:
    case IS_STRING:
    case IS_DOUBLE:
    case IS_LONG:
    case IS_RESOURCE:
		php_error(E_ERROR, "get_ABSTRACT_MAP_func_impl: while iterating loop: got scalar value instead of associative array");
		return NULL;
	/* object and array are both supported in get_ABSTRACT_VALUE_func_impl */
    case IS_OBJECT:
    case IS_ARRAY:
		return (ABSTRACT_MAP*) zvalue;
    default:
		php_error(E_ERROR, "get_ABSTRACT_MAP_func_impl: while iterating loop: value has unsupported php type %d", (Z_TYPE_P(zvalue)));
	}
	return NULL;
}

static int get_ABSTRACT_ARRAY_length_func_impl (ABSTRACT_DATASTATE* ze_self, ABSTRACT_ARRAY* arr) {
	int ztype = Z_TYPE_P((zval*)arr);
	if (ztype==IS_ARRAY) return zend_hash_num_elements(Z_ARRVAL_P((zval*)arr));
	if (ztype == IS_OBJECT) {
		zval* zretval = NULL;
		if (test_and_call_zend_method_with_0_params((zval*)arr, ((struct _ze_htp_object*) ze_self)->method_loop_count, ((struct _ze_htp_object*) ze_self)->method_loop_count_len, &zretval)) {
			if (zretval != NULL) return Z_LVAL_P(zretval);
		}
	}
	return -1;
}

static const char* const STRUE ="1";
static const char* const SFALSE ="0";
static PSTRING pstring_true (void) {
	PSTRING ret = {STRUE, STRUE+1};
	return ret;
}
static PSTRING pstring_false (void) {
	PSTRING ret = {SFALSE, SFALSE+1};
	return ret;
}

static int is_ABSTRACT_VALUE_true_func_impl (ABSTRACT_DATASTATE* none, ABSTRACT_VALUE* val) {
	return zend_is_true((zval*) val);
}

// memory leaks during session: zstringval
static PSTRING ABSTRACT_VALUE2PSTRING_func_impl (ABSTRACT_DATASTATE* none, ABSTRACT_VALUE* val) {
	PSTRING pstr = {NULL, NULL};
	zend_class_entry *ce;
	zval* zabstrval = (zval*) val;
	zval* zstringval=NULL;
	switch (Z_TYPE_P(zabstrval)) {
    case IS_NULL:
		return pstr;
    case IS_BOOL:
        return Z_LVAL_P(zabstrval) ? pstring_true(): pstring_false();
    case IS_STRING:
        pstr.begin = Z_STRVAL_P(zabstrval);
        pstr.endnext = pstr.begin + Z_STRLEN_P(zabstrval);
		return pstr;
    case IS_DOUBLE:
    case IS_LONG:
		SEPARATE_ZVAL(&zabstrval);
		convert_to_string(zabstrval);
        pstr.begin = Z_STRVAL_P(zabstrval);
        pstr.endnext = pstr.begin + Z_STRLEN_P(zabstrval);
		return pstr;
    case IS_OBJECT:
		/* call __toString */
		ce = Z_OBJCE_P(zabstrval);
		if (! ce->__tostring) {
			php_error(E_RECOVERABLE_ERROR, "Method %s::__toString() not found!", ce->name);
		} else {
#if (PHP_MAJOR_VERSION < 5) || (PHP_MAJOR_VERSION == 5 && PHP_MINOR_VERSION < 3) 
			zend_call_method(&zabstrval, ce, &ce->__tostring, "__tostring", sizeof("__tostring")-1, &zstringval, 0, NULL, NULL TSRMLS_CC);
			if (zstringval!=NULL || EG(exception)) {
#else
			if (zend_call_method_with_0_params(&zabstrval, ce, &ce->__tostring, "__tostring", &zstringval) || EG(exception)) {
#endif
				if (EG(exception)) {
					if (zstringval) {
						zval_ptr_dtor(&zstringval);
					}
					php_error(E_RECOVERABLE_ERROR, "Method %s::__toString() must not throw an exception", ce->name);
				} else if (Z_TYPE_P(zstringval) == IS_STRING) {
					pstr.begin = Z_STRVAL_P(zstringval);
					pstr.endnext = pstr.begin + Z_STRLEN_P(zstringval);
				} else {
					php_error(E_RECOVERABLE_ERROR, "Method %s::__toString() must return a string value", ce->name);
				}
			}
		}
		return pstr;
    case IS_RESOURCE:
    case IS_ARRAY:
		php_error(E_RECOVERABLE_ERROR, "Can't evaluate a Resource or an Array to a string value");
		return pstr;
    default:
		return pstr;
    }
}

static void writer_func_impl (ABSTRACT_WRITER* none, const char* begin, const char* endnext) {
	TSRMLS_FETCH();
	php_write((void *)begin, endnext-begin TSRMLS_CC);
}

struct userfunc_args_header {
  size_t capacity;
  size_t len;
  zval** zargv;
};

#define USERFUNC_ARGS_INITIAL_CAPACITY 32

ABSTRACT_ARGLIST*  init_expr_arglist_func_impl (ABSTRACT_CALLER* none) {
	struct userfunc_args_header* userfunc_args = emalloc (sizeof (struct userfunc_args_header));
	userfunc_args->capacity = USERFUNC_ARGS_INITIAL_CAPACITY;
	userfunc_args->len = 0;
	userfunc_args->zargv = emalloc(USERFUNC_ARGS_INITIAL_CAPACITY * sizeof userfunc_args->zargv);
	return (ABSTRACT_ARGLIST*) userfunc_args;
}

void  free_expr_arglist_func_impl (ABSTRACT_ARGLIST* userfunc_args) {
	size_t i;
	const size_t len = ((struct userfunc_args_header*)userfunc_args)->len;
	zval** zargv = ((struct userfunc_args_header*)userfunc_args)->zargv;
	for (i=0; i< len; i++) {zval_dtor(zargv[i]);}
	efree(zargv);
	efree(userfunc_args);
}

void  push_expr_arglist_func_impl (ABSTRACT_ARGLIST* args, ABSTRACT_EXPRVAL* e) {
	zval* ze;
	PSTRING pstr;
	struct userfunc_args_header* userfunc_args = (struct userfunc_args_header*) args;
	if (++(userfunc_args->len) >= userfunc_args->capacity) {
		userfunc_args->capacity *= 2;
		userfunc_args->zargv = erealloc(userfunc_args->zargv, userfunc_args->capacity * sizeof userfunc_args->zargv);
	}
	ALLOC_INIT_ZVAL(ze);
	switch (tmplpro_get_expr_type (e)) {
    case EXPR_TYPE_NULL: 
		; break;
	case EXPR_TYPE_INT: 
		ZVAL_LONG (ze, (long)tmplpro_get_expr_as_int64(e));	; break;
	case EXPR_TYPE_DBL:
		ZVAL_DOUBLE (ze, tmplpro_get_expr_as_double(e));	; break;
	case EXPR_TYPE_PSTR:
		pstr = tmplpro_get_expr_as_pstring(e);
		/* dup=1 cause a small ememory leak until the end of request,
		   but dup = 0 might cause problems if userfunc store/modify its value 
		   or has other side effects. So let us rely on efree on rshutdown. */
		ZVAL_STRINGL(ze, (char*)pstr.begin,pstr.endnext-pstr.begin,1);
		; break;
    default: ; break;
	}
	userfunc_args->zargv[userfunc_args->len-1] = ze;
}

void  call_expr_userfnc_func_impl (ABSTRACT_CALLER* caller, ABSTRACT_ARGLIST* args, ABSTRACT_USERFUNC* zfunc, ABSTRACT_EXPRVAL* retval) {
	ze_htp_object *ze_obj = (ze_htp_object *) caller;
	struct userfunc_args_header* userfunc_args = (struct userfunc_args_header*) args;
	char* name;
	int argc = userfunc_args->len;
	zval **params = userfunc_args->zargv;
	zval zretval;
	TSRMLS_FETCH();
	//php_printf("call_expr_userfnc_func_impl: is going to be called with %d args\n", argc);

	/* note: zfunc is already checked to be string/array in is_expr_userfnc_func_impl */
	if (!zend_is_callable((zval*) zfunc, 0, &name TSRMLS_CC)) {
		php_error_docref1(NULL TSRMLS_CC, name, E_WARNING, "name of a user function is invalid");
	} else if (call_user_function(EG(function_table), NULL, (zval*) zfunc, &zretval, argc, params TSRMLS_CC) == SUCCESS) {
		PSTRING pstr = {NULL, NULL};
		switch (Z_TYPE(zretval)) {
		case IS_NULL:
			tmplpro_set_expr_as_null(retval); break;
		case IS_BOOL:
		case IS_LONG:
			tmplpro_set_expr_as_int64(retval, Z_LVAL(zretval)); break;
		case IS_DOUBLE:
			tmplpro_set_expr_as_double(retval, Z_DVAL(zretval)); break;
		case IS_STRING:
			pstr.begin = Z_STRVAL(zretval);
			pstr.endnext = pstr.begin + Z_STRLEN(zretval);
			tmplpro_set_expr_as_pstring(retval, pstr); break;
		case IS_RESOURCE:
		case IS_ARRAY:
		case IS_OBJECT:
		default:
			php_error_docref1(NULL TSRMLS_CC, name, E_WARNING, "user function returned something complicated.");
		}
		
	} else {
		/* NOTE: convert_to_string is safe, as we constructed those zvals themselves */
		if (argc > 1) {
			convert_to_string(params[1]);
			if (argc > 2) {
				convert_to_string(params[2]);
				php_error_docref1(NULL TSRMLS_CC, name, E_WARNING, "Unable to call %s(%s,%s)", name, Z_STRVAL_P(params[1]), Z_STRVAL_P(params[2]));
			} else {
				php_error_docref1(NULL TSRMLS_CC, name, E_WARNING, "Unable to call %s(%s)", name, Z_STRVAL_P(params[1]));
			}
		} else {
			php_error_docref1(NULL TSRMLS_CC, name, E_WARNING, "Unable to call %s()", name);
		}
	}
	efree(name);
}


ABSTRACT_USERFUNC* is_expr_userfnc_func_impl (ABSTRACT_FUNCMAP* fmap, PSTRING name) {
	zval **mapvalptr;
	char* tmpkey;
	size_t keylen = name.endnext-name.begin;
	if (NULL== fmap) return NULL;
	tmpkey = estrndup(name.begin, keylen);
	if(zend_hash_find((HashTable *) fmap, (char*)tmpkey, keylen+ COUNT_OF_END_STRING_ZERO, 
					  (void **)&mapvalptr) == FAILURE || mapvalptr == NULL || *mapvalptr==NULL) {
		efree(tmpkey);
		return NULL;
	}
	efree(tmpkey);
	if (Z_TYPE_PP(mapvalptr) != IS_STRING && Z_TYPE_PP(mapvalptr) != IS_ARRAY) {
		SEPARATE_ZVAL(mapvalptr);
		convert_to_string_ex(mapvalptr);
	}
	//php_printf("is_expr_userfnc_func_impl: true for %.*s\n", (int)(name.endnext - name.begin),name.begin);

	return (ABSTRACT_USERFUNC*) *mapvalptr;
}

/* ==================================================================== */

static zend_object_handlers htmltmplpro_object_handlers;

/* {{{ htmltmplpro_class_functions */
static zend_function_entry htmltmplpro_class_functions[] = {
	PHP_ME(htmltmplpro,	  	getopt, NULL,0)
	PHP_ME(htmltmplpro,	  	setopt, NULL,0)
	PHP_ME(htmltmplpro,	  	output, NULL,0)
	PHP_ME(htmltmplpro,	  	print, NULL,0)
	PHP_ME(htmltmplpro,	  	reset, NULL,0)
	PHP_ME_MAPPING(setInputFile,	htmltmplpro_set_input_file, 	NULL, 0)
	PHP_ME_MAPPING(setInputString,	htmltmplpro_set_input_string, 	NULL, 0)
	PHP_ME_MAPPING(getInputFile,	htmltmplpro_get_input_file, 	NULL, 0)
	PHP_ME_MAPPING(getInputString,	htmltmplpro_get_input_string, 	NULL, 0)
	PHP_ME_MAPPING(error,	htmltmplpro_last_error, 	NULL, 0)
	{NULL, NULL, NULL}
};
/* }}} */

/* {{{ htp_object_free_storage */
static void htp_object_free_storage(void *object TSRMLS_DC)
{
	ze_htp_object * ze_obj = (ze_htp_object *) object;
	//php_printf("htp_object_free_storage\n");

	if (!ze_obj) {
		return;
	}
	if (ze_obj->htp_ptr) {
		tmplpro_param_free(ze_obj->htp_ptr);
	}
	ze_obj->htp_ptr = NULL;

	HTP_ZVAL_FIELD_CLEAR(ze_obj, input_zval);
	HTP_ZVAL_FIELD_CLEAR(ze_obj, userfuncmap);
	HTP_ZVAL_FIELD_CLEAR(ze_obj, path_zarray_arg);
	_free_option_path(ze_obj);

	zend_object_std_dtor(&ze_obj->zo TSRMLS_CC);
	efree(ze_obj);
}
/* }}} */

#define ADD_CONST_INT(NAME,VALUE) \
    constval = pemalloc(sizeof(zval), 1); \
    INIT_PZVAL(constval); \
    ZVAL_LONG(constval, VALUE); \
    zend_hash_add(&ce->constants_table, NAME, sizeof(NAME), (void*)&constval, sizeof(zval*), NULL);


void php_htmltmplpro_register_constants(zend_class_entry *ce) { 
    zval *constval; 
	ADD_CONST_INT("ASK_NAME_DEFAULT",	ASK_NAME_DEFAULT);
	ADD_CONST_INT("ASK_NAME_AS_IS",		ASK_NAME_AS_IS);
	ADD_CONST_INT("ASK_NAME_LOWERCASE",	ASK_NAME_LOWERCASE);
	ADD_CONST_INT("ASK_NAME_UPPERCASE",	ASK_NAME_UPPERCASE);
	ADD_CONST_INT("ASK_NAME_MASK",		ASK_NAME_MASK);

	ADD_CONST_INT("OPT_ESCAPE_NO",	HTML_TEMPLATE_OPT_ESCAPE_NO);
	ADD_CONST_INT("OPT_ESCAPE_HTML",HTML_TEMPLATE_OPT_ESCAPE_HTML);
	ADD_CONST_INT("OPT_ESCAPE_URL",	HTML_TEMPLATE_OPT_ESCAPE_URL);
	ADD_CONST_INT("OPT_ESCAPE_JS",	HTML_TEMPLATE_OPT_ESCAPE_JS);
} 



/* {{{ htp_object_new */
static zend_object_value htp_object_new(zend_class_entry *class_type TSRMLS_DC)
{
	ze_htp_object *intern;
	zval *tmp;
	zend_object_value retval;
	struct tmplpro_param* param = tmplpro_param_init();
	
	intern = (ze_htp_object*) emalloc(sizeof(ze_htp_object));
	memset(&intern->zo, 0, sizeof(zend_object));
	intern->htp_ptr = param;
	intern->input_zval = NULL;
	intern->path = NULL;
	intern->userfuncmap = NULL;
	intern->path_zarray_arg = NULL;
	intern->method_loop_init = "loop_init";
	intern->method_loop_init_len = sizeof("loop_init")-1;
	intern->method_loop_next = "loop_next";
	intern->method_loop_next_len = sizeof("loop_next")-1;
	// the Countable interface
	intern->method_loop_count = "count";
	intern->method_loop_count_len = sizeof("count")-1;

	tmplpro_set_option_WriterFuncPtr(param, writer_func_impl);
	tmplpro_set_option_GetAbstractValFuncPtr(param, get_ABSTRACT_VALUE_func_impl);
	tmplpro_set_option_AbstractVal2pstringFuncPtr(param, ABSTRACT_VALUE2PSTRING_func_impl);
	tmplpro_set_option_AbstractVal2abstractArrayFuncPtr(param, ABSTRACT_VALUE2ABSTRACT_ARRAY_func_impl);
	tmplpro_set_option_GetAbstractMapFuncPtr(param, get_ABSTRACT_MAP_func_impl);
	tmplpro_set_option_IsAbstractValTrueFuncPtr(param, is_ABSTRACT_VALUE_true_func_impl);
	tmplpro_set_option_GetAbstractArrayLengthFuncPtr(param, get_ABSTRACT_ARRAY_length_func_impl);

	tmplpro_set_option_InitExprArglistFuncPtr(param, init_expr_arglist_func_impl);
	tmplpro_set_option_FreeExprArglistFuncPtr(param, free_expr_arglist_func_impl);
	tmplpro_set_option_PushExprArglistFuncPtr(param, push_expr_arglist_func_impl);
	tmplpro_set_option_CallExprUserfncFuncPtr(param, call_expr_userfnc_func_impl);
	tmplpro_set_option_IsExprUserfncFuncPtr(param, is_expr_userfnc_func_impl);

	tmplpro_set_option_ext_calluserfunc_state(param, intern);
	tmplpro_set_option_ext_data_state(param, intern);
	//tmplpro_set_option_(param, );

	zend_object_std_init(&intern->zo, class_type TSRMLS_CC);
	zend_hash_copy(intern->zo.properties, &class_type->default_properties, (copy_ctor_func_t) zval_add_ref,
					(void *) &tmp, sizeof(zval *));

	retval.handle = zend_objects_store_put(intern,
						/* NULL, /* TODO: should? be not NULL, but */
								 (zend_objects_store_dtor_t)zend_objects_destroy_object, 
						(zend_objects_free_object_storage_t) htp_object_free_storage,
						NULL TSRMLS_CC);
	
	retval.handlers = (zend_object_handlers *) & htmltmplpro_object_handlers;
	return retval;
}
/* }}} */

/* {{{ PHP_MINIT_FUNCTION
 */
PHP_MINIT_FUNCTION(htmltmplpro)
{
	zend_class_entry ce;
	/* register in php */
	memcpy(&htmltmplpro_object_handlers, zend_get_std_object_handlers(), sizeof(zend_object_handlers));
	htmltmplpro_object_handlers.clone_obj      = NULL;
	INIT_CLASS_ENTRY(ce, "TemplateEngine", htmltmplpro_class_functions);
	ce.create_object = htp_object_new;
	htmltmplpro_class_entry_ce = zend_register_internal_class(&ce TSRMLS_CC);
	php_htmltmplpro_register_constants(htmltmplpro_class_entry_ce);

	/* initialize library */
	tmplpro_procore_init();
	return SUCCESS;
}
/* }}} */

/*---------------------------------------------------*/













/* Remove the following function when you have succesfully modified config.m4
   so that your module can be compiled into PHP, it exists only for testing
   purposes. */

/* Every user-visible function in PHP should document itself in the source */
/* {{{ proto string htmltmplpro_versioninfo(void)
   Return a version string */
static PHP_FUNCTION(htmltmplpro_versioninfo)
{
	int len;
	char *strg;
	len = spprintf(&strg, 0, "%s php extension version %s, using C htmltmplpro library version %s, compiled for PHP version %s", 
				   HTMLTMPLPRO_EXT_NAME, 
				   HTMLTMPLPRO_EXT_VERSION,
				   tmplpro_version(),
				   PHP_VERSION);
	RETURN_STRINGL(strg, len, 0);
}
/* }}} */
/* The previous line is meant for vim and emacs, so it can correctly fold and 
   unfold functions in source code. See the corresponding marks just before 
   function definition, where the functions purpose is also documented. Please 
   follow this convention for the convenience of others editing your code.
*/


/*
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * End:
 * vim600: noet sw=4 ts=4 fdm=marker
 * vim<600: noet sw=4 ts=4
 */
